/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 */
#include "wanpipe_lip_lapb.h"

/*
 *	State machine for state 0, Disconnected State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state0_machine(lapb_cb *lapb, void *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:

			if (wpabs_test_bit(LAPB_DOWN,&lapb->critical)){
				return;
			}
			
			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_SABME:

			if (wpabs_test_bit(LAPB_DOWN,&lapb->critical)){
				return;
			}

			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			} else {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			break;

		default:
			break;
	}

	wpabs_skb_free(skb);
}

/*
 *	State machine for state 1, Awaiting Connection State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state1_machine(lapb_cb *lapb, void *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_SABME:
			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			} else {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
			lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_UA:
			if (frame->pf) {
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_DM:
			if (frame->pf) {
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_indication(lapb, LAPB_REFUSED);
			}
			break;

		default:
			break;
	}

	wpabs_skb_free(skb);
}

/*
 *	State machine for state 2, Awaiting Release State.
 *	The handling of the timer(s) is in file lapb_timer.c
 */
static void lapb_state2_machine(lapb_cb *lapb, void *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
		case LAPB_SABME:
			lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_DISC:
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_UA:
			if (frame->pf) {
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_DM:
			if (frame->pf) {
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_indication(lapb, LAPB_NOTCONNECTED);
			}
			break;

		case LAPB_I:
		case LAPB_REJ:
		case LAPB_RNR:
		case LAPB_RR:
			if (frame->pf){
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;
				
		default:
			break;
	}

	wpabs_skb_free(skb);
}





/*
 *	State machine for state 3, Connected State.
 *	The handling of the timer(s) is in file lapb_timer.c
 */
static void lapb_state3_machine(lapb_cb *lapb, void *skb, struct lapb_frame *frame)
{
	int queued = 0;
	int modulus;
	
	modulus = (lapb->mode & LAPB_EXTENDED) ? LAPB_EMODULUS : LAPB_SMODULUS;

	lapb->nr_type=frame->type;

	switch (frame->type) {
		case LAPB_SABM:
			
			wpabs_debug_event( "%s: LAPB Event : State Connected : Rx SABM !\n",lapb->name);
			
			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
				lapb_clear_queues(lapb);
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_requeue_frames(lapb);
			}

			lapb_connect_indication(lapb, LAPB_OK);
			break;

		case LAPB_SABME:
			
			wpabs_debug_event( "%s: LAPB Event : State Connected : Rx SABME !\n",lapb->name);

			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_requeue_frames(lapb);
			} else {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
			wpabs_debug_event( "%s: LAPB Event : State connected : Rx DISC\n",lapb->name);	
			lapb_clear_queues(lapb);
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
			lapb_disconnect_indication(lapb, LAPB_OK);
			break;

		case LAPB_DM:
			wpabs_debug_event( "%s: LAPB Event : State connected : Rx DM\n",lapb->name);
			lapb_clear_queues(lapb);
			lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_disconnect_indication(lapb, LAPB_NOTCONNECTED);
			break;

		case LAPB_RNR:

			wpabs_debug_event( "%s: LAPB Event : State connected : Rx RNR\n",lapb->name);

			lapb->condition |= LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);

			if (lapb_validate_nr(lapb, frame->nr)) {
				lapb_check_iframes_acked(lapb, frame->nr);
			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_RR:
			lapb->condition &= ~LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);
			if (lapb_validate_nr(lapb, frame->nr)) {
				lapb_check_iframes_acked(lapb, frame->nr);
			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_REJ:
			
			wpabs_debug_event( "%s: LAPB Event : State Connected : Rx REJ\n",lapb->name);
			lapb->condition &= ~LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);

			if (lapb_validate_nr(lapb, frame->nr)) {

				lapb_frames_acked(lapb, frame->nr);
				lapb_stop_t1timer(lapb);
				lapb->n2count = 0;
				lapb_requeue_frames(lapb);

			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_I:
			if (!lapb_validate_nr(lapb, frame->nr)) {

				lapb->old_nr_type=frame->type;

				wpabs_debug_event( "%s: Lapb Error: I-Frame Rx invalid NR count %d !!!\n",
						lapb->name,frame->nr);
				
				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
				break;
			}

			if (lapb->condition & LAPB_PEER_RX_BUSY_CONDITION){
				lapb_frames_acked(lapb, frame->nr);
			}else{
				lapb_check_iframes_acked(lapb, frame->nr);
			}
			
			if (frame->ns == lapb->vr){

				void *newskb;
				
				/*
				 * If upper layer has dropped the frame, we
				 * basically ignore any further protocol
				 * processing. This will cause the peer
				 * to re-transmit the frame later like
				 * a frame lost on the wire.
				 */

				if (wpabs_skb_queue_len(lapb->rx_queue) >= lapb->window){
					wpabs_debug_test("%s: Rx window closed dropping packet!\n",
							lapb->name);

					break;
				}

				queued = 1;	

				wpabs_skb_queue_tail(lapb->rx_queue,skb);
				lapb->vr = (lapb->vr + 1) % modulus;

				while ((newskb=wpabs_skb_dequeue(lapb->rx_queue)) != NULL){
					if (lapb_data_indication(lapb, newskb) != 0){
						wpabs_debug_test( "%s: Failed to pass data up : top layer busy!\n",
								lapb->name);
						wpabs_skb_queue_head(lapb->rx_queue,newskb);
						break;
					}else{
						lapb->condition &= ~LAPB_RNR_SENT_CONDITION;

					}
				}
				
				lapb->condition &= ~LAPB_REJECT_CONDITION;
				
				if (frame->pf){
					lapb_enquiry_response(lapb);
				}else{
					if (!(lapb->condition & LAPB_ACK_PENDING_CONDITION)){
					
						lapb->condition |= LAPB_ACK_PENDING_CONDITION;

						/* If the rx queue has reached maximum, send
						 * out an RNR to flow control the incoming
						 * traffic, until the queue is offloaded by
						 * the above protocol */
						if (wpabs_skb_queue_len(lapb->rx_queue) >= lapb->window){
							lapb->condition |= LAPB_RNR_SENT_CONDITION; 
#if 0
							wpabs_debug_event("%s: LAPB receive queue limit! Sending RNR!\n",
									lapb->name);
#endif
							lapb_send_control(lapb, LAPB_RNR, 
							        LAPB_POLLOFF, LAPB_RESPONSE);
							break;
						}

						/* If the T2 timer is zero, then send the
						 * RR response to the I frame right away,
						 * otherwise start the T2 timer and send
						 * the response after timer expires 
						 *
						 * NC: T2==0 -> send RR after every frame!
						 *     However, don't send RR if there
						 *     are I-frames pending.  RR is an
						 *     overkill.
						 * */
						if (!lapb->t2){
							if (!wpabs_skb_queue_len(lapb->write_queue) &&
							    !wpabs_skb_queue_len(lapb->async_queue)){
								lapb_send_control(lapb, LAPB_RR, 
									  LAPB_POLLOFF, LAPB_RESPONSE);
							}else{
#if 0
								DEBUG_TEST("Lapb Q WQ=%d AQ=%d\n",
										wpabs_skb_queue_len(lapb->write_queue),
										wpabs_skb_queue_len(lapb->async_queue));
#endif
							}
							lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
						}else{
							lapb_start_t2timer(lapb);
						}
					}
				}
			} else {
				if (lapb->condition & LAPB_REJECT_CONDITION) {
					if (frame->pf)
						lapb_enquiry_response(lapb);
				} else {
#if 0
					DEBUG_TEST("%s: Error: Lapb sending reject: In Ns=%d != Orig Vr=%d\n",
							lapb->name,frame->ns,lapb->vr);

#endif
					lapb->condition |= LAPB_REJECT_CONDITION;
					lapb_send_control(lapb, LAPB_REJ, frame->pf, LAPB_RESPONSE);
					lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
				}
			}
			break;

		case LAPB_FRMR:

			wpabs_debug_event( "%s: LAPB Event : State Connected : Rx FRMR\n",lapb->name);
			lapb_establish_data_link(lapb);
	
			lapb_requeue_frames(lapb);
			lapb_state_change(lapb,LAPB_STATE_CONNECTING);
			break;

		case LAPB_ILLEGAL:

			lapb->frmr_data = *frame;
			lapb->frmr_type = LAPB_FRMR_W;
			lapb_transmit_frmr(lapb);
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
			lapb->n2count = 0;
			break;

		default:
			break;
	}

	if (!queued){
		wpabs_skb_free(skb);
	}
}

/*
 *	State machine for state 4, Frame Reject State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state4_machine(lapb_cb *lapb, void *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
			if (lapb->mode & LAPB_EXTENDED) {

				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {


				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_SABME:

			if (lapb->mode & LAPB_EXTENDED) {


				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			} else {

				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		default:
			break;
	}

	wpabs_skb_free(skb);
}


static void lapb_rx_recover(lapb_cb *lapb)
{
	void *newskb;
	int rx_ok=0;

	if (!wpabs_skb_queue_len(lapb->rx_queue)){
		return;
	}
	
	while ((newskb=wpabs_skb_dequeue(lapb->rx_queue)) != NULL){
		if (lapb_data_indication(lapb, newskb) != 0){
			wpabs_debug_test( "%s: Failed to pass data up : top layer busy!\n",
					lapb->name);
			wpabs_skb_queue_head(lapb->rx_queue,newskb);
			break;
		}
		rx_ok=1;
	}

	if (!rx_ok){
		return;
	}
	
	if (wpabs_skb_queue_len(lapb->rx_queue) < lapb->window){
		if (lapb->condition & LAPB_RNR_SENT_CONDITION){
			wpabs_debug_event("%s: %s() Clearing RNR CONDITION!!!\n",
					lapb->name,__FUNCTION__);
			lapb->condition &= ~LAPB_RNR_SENT_CONDITION;
		}
	}
	
	return;
}


/*
 *	Process an incoming LAPB frame
 */
void lapb_data_input(lapb_cb *lapb, void *skb)
{
	struct lapb_frame frame;
	int len = wpabs_skb_len(skb);

	FUNC_BEGIN();
	
	if (wpabs_test_bit(LAPB_UNREG_CRIT,&lapb->critical)){
		wpabs_skb_free(skb);
		return;
	}

	if (wpabs_test_and_set_bit(LAPB_RX_SMP,&lapb->critical)){
		wpabs_debug_event( "---!!!__ Critical in Lapb Rx!\n");
	}
	
	lapb_decode(lapb, skb, &frame);

	if (frame.type == LAPB_CORRUPT){
		wpabs_skb_free(skb);
		goto lapb_data_input_error;
	}

	if (wpabs_skb_len(skb) > lapb->mtu){
		wpabs_debug_event( "LAPB RX: incoming packet size %d exceed maximum %d\n",wpabs_skb_len(skb), lapb->mtu);
		frame.type = LAPB_ILLEGAL;
	}
	lapb_update_stats(lapb,frame.type,len,0);

	/* When we receive a valid frame, restart
	 * the T4 timer. Used for link inactivity
	 * detection */
	if (frame.type != LAPB_ILLEGAL && 
	    !wpabs_test_bit(LAPB_DOWN,&lapb->critical)){
		lapb_start_t4timer(lapb, 1);
	}

	switch (lapb->state) {
		case LAPB_STATE_DISCONNECTED:
			lapb_state0_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_CONNECTING:
			lapb_state1_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_DISCONNECTING:
			lapb_state2_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_CONNECTED:
			lapb_state3_machine(lapb, skb, &frame);
			lapb_rx_recover(lapb);
			break;
		case LAPB_STATE_FRAME_REJECT:
			lapb_state4_machine(lapb, skb, &frame);
			break;
		default:
			wpabs_debug_event( "%s: Lapb Rx in invalid state: droping packet\n",
					lapb->name);
			wpabs_skb_free(skb);
			break;
	}

lapb_data_input_error:
	wpabs_clear_bit(LAPB_RX_SMP,&lapb->critical);

	FUNC_END();
}

